@@ -1,14 +1,17 @@
 import json
 import urllib.parse
+import io
 import difflib
 import inspect
 from contextvars import copy_context
+from contextlib import redirect_stdout
 
 import fiona
 from tqdm import tqdm
 import geopandas as gpd
 from shapely.ops import transform
 from shapely.geometry import Polygon, MultiPolygon, LineString, Point, box
+from shapely.geometry.collection import GeometryCollection
 from shapely.geometry.base import BaseGeometry
 import matplotlib.pyplot as plt
 import matplotlib.colors as mcolors
@@ -135,14 +138,25 @@ class Mundi:
         # TODO: drop duplicates, except it's very slow
         #.drop_duplicates(subset=['geometry'])
         Q = MundiQ(self.main.geometry_collection('EPSG:4326')[element_index], self.mapdata, plot_target=('geojson' if output_type == 'geojson' else ax), units=self.units, clip_distance=clip_distance)
-        res = Q.call_process(fn)
+
+        with redirect_stdout(io.StringIO()) as f:
+            res = Q.call_process(fn)
+        # add stdout
+        res['_stdout'] = f.getvalue()
+        res['_id'] = element_index
 
         if output_type == 'geojson':
-            # convert to dfs
-            dfs = gpd.GeoDataFrame(data=Q.plot_contents, crs='EPSG:4326', columns=['geometry', 'fill'], geometry='geometry')
+            # merge geometries into one
+            geom_col = GeometryCollection([r[0] for r in Q.plot_contents])
 
-            # sanitize object before dumping to json
-            return json.dumps(dfs.__geo_interface__)
+            return json.dumps({
+                "type": "FeatureCollection",
+                "features": [{
+                    "type": "GeometryCollection",
+                    "geometries": geom_col.__geo_interface__['geometries'],
+                    "properties": { k: (int(v) if isinstance(v, int) else (float(v) if isinstance(v, float) else str(v))) for (k, v) in res.items() if not isinstance(v, BaseGeometry)}
+                }]
+            })
         elif output_type == 'matplotlib':
             ax.legend(handles=Q.plot_handles, loc='upper right')
             plt.show()
@@ -164,9 +178,16 @@ class Mundi:
         for idx, original_shape in finiter:
             # TODO fn(Q) can edit window
 
+            user_printed = None
+
             try:
                 Q = MundiQ(original_shape, self.mapdata)
-                res = Q.call_process(fn)
+
+                # capture stdout
+                with redirect_stdout(io.StringIO()) as f:
+                    res = Q.call_process(fn)
+                user_printed = f.getvalue()
+                
             except NoProjectionFoundError:
                 continue
 
@@ -178,6 +199,8 @@ class Mundi:
             if not isinstance(res, dict):
                 raise TypeError('function passed to mundi.q() must return dict or None but instead got %s' % type(res).__name__)
 
+            # type check that keys are always the same
+            # ignores _stdout and _id
             if res_keys is None:
                 res_keys = res.keys()
 
@@ -189,6 +212,10 @@ class Mundi:
 
                 if res_shapely_col == 'geometry':
                     res_outs['geometry'] = []
+
+                # internal columns
+                res_outs['_stdout'] = []
+                res_outs['_id'] = []
             elif res_keys != res.keys():
                 raise TypeError('function passed to mundi.q() returned dict with different keys')
 
@@ -207,6 +234,9 @@ class Mundi:
             if res_shapely_col == 'geometry':
                 res_outs['geometry'].append(original_shape)
 
+            res_outs['_stdout'].append(user_printed)
+            res_outs['_id'].append(idx)
+
         # if res_outs is empty, give a useful error message
         # creating a GeoDataFrame with an empty array gives an error
         if len(res_outs) == 0:
